from __future__ import annotations
import asyncio
import pandas as pd
from datetime import datetime, timedelta, timezone
from aiohttp import ClientSession, ClientError
import hashlib
import hmac
import base64
import re
from enum import Enum
from http import HTTPStatus
import json
from typing import Any, Optional

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from quartz_solar_forecast.inverters.inverter import AbstractInverter

try:
    import async_timeout
except:
    print('Could not import `async_timeout`')


# VERSION
RESOURCE_PREFIX = '/v1/api/'

VERB = "POST"

# Endpoints
INVERTER_LIST = RESOURCE_PREFIX + 'inverterList'
INVERTER_DAY = RESOURCE_PREFIX + 'inverterDay'


class SolisSettings(BaseSettings):
    model_config = SettingsConfigDict(env_file='.env', extra='ignore')

    api_url: str = Field(alias="SOLIS_CLOUD_API_URL", default='https://www.soliscloud.com')
    port: str = Field(alias="SOLIS_CLOUD_API_PORT", default='13333')
    api_key: str = Field(alias="SOLIS_CLOUD_API_KEY")
    client_secret: str = Field(alias="SOLIS_CLOUD_API_KEY_SECRET")


class SolisInverter(AbstractInverter):
    def __init__(self, settings: SolisSettings):
        self.__settings = settings

    def get_data(self, ts: pd.Timestamp) -> Optional[pd.DataFrame]:
        try:
            return asyncio.run(get_solis_data(self.__settings))
        except Exception as e:
            print(f"Error retrieving Solis data: {str(e)}")
            return None

class SoliscloudAPI():
    """Class with functions for reading data from the Soliscloud Portal."""

    class SolisCloudError(Exception):
        """
        Exception raised for timeouts during calls.
        """

        def __init__(self, message="SolisCloud API error"):

            self.message = message
            super().__init__(self.message)

    class HttpError(SolisCloudError):
        """
        Exception raised for HTTP errors during calls.
        """

        def __init__(self, statuscode, message=None):
            self.statuscode = statuscode
            self.message = message
            if not message:
                if statuscode == 408:
                    now = datetime.now().strftime("%d-%m-%Y %H:%M GMT")
                    self.message = f"Your system time is different from server time, your time is {now}"
                else:
                    self.message = f"Http status code: {statuscode}"
            super().__init__(self.message)

    class TimeoutError(SolisCloudError):
        """
        Exception raised for timeouts during calls.
        """

        def __init__(self, message="Timeout error occurred"):

            self.message = message
            super().__init__(self.message)

    class ApiError(SolisCloudError):
        """
        Exception raised for errors during API calls.
        """

        def __init__(self, message="Undefined API error occurred", code="Unknown", response=None):

            self.message = message
            self.code = code
            self.response = response
            super().__init__(self.message)

        def __str__(self):
            return f'API returned an error: {self.message}, error code: {self.code}, response: {self.response}'

    def __init__(self, domain: str, session: ClientSession) -> None:
        self._domain = domain.rstrip("/")
        self._session: ClientSession = session

    class DateFormat(Enum):
        DAY = 0
        MONTH = 1
        YEAR = 2

    @property
    def domain(self) -> str:
        """ Domain name."""
        return self._domain

    @property
    def session(self) -> ClientSession:
        """ aiohttp client session ID."""
        return self._session

    # All methods take key and secret as positional arguments followed by
    # one or more keyword arguments

    async def inverter_list(self, key_id: str, secret: bytes, /, *,
        page_no: int = 1,
        page_size: int = 20,
        station_id: str = None,
        nmi_code: str = None
    ) -> dict[str, str]:
        """Inverter list"""

        if page_size > 100:
            raise SoliscloudAPI.SolisCloudError("PageSize must be <= 100")

        params: dict[str, Any] = {'pageNo': page_no, 'pageSize': page_size}
        if station_id is not None:
            # If not specified all inverters for all stations for key_id are returned
            params['stationId'] = station_id
        if nmi_code is not None:
            params['nmiCode'] = nmi_code
        return await self._get_records(INVERTER_LIST, key_id, secret, params)

    async def inverter_day(self, key_id: str, secret: bytes, /, *,
        currency: str,
        time: str,
        time_zone: int,
        inverter_id: int = None,
        inverter_sn: str = None
    ) -> dict[str, str]:
        """Inverter daily graph"""

        SoliscloudAPI._verify_date(SoliscloudAPI.DateFormat.DAY, time)
        params: dict[str, Any] = {'money': currency, 'time': time, 'timeZone': time_zone}

        if (inverter_id is not None and inverter_sn is None):
            params['id'] = inverter_id
        elif (inverter_id is None and inverter_sn is not None):
            params['sn'] = inverter_sn
        else:
            raise SoliscloudAPI.SolisCloudError("Only pass one of inverter_id or inverter_sn \
                as identifier")

        return await self._get_data(INVERTER_DAY, key_id, secret, params)

    async def _get_records(self, canonicalized_resource: str, key_id: str, secret: bytes, params: dict[str, Any]):
        """
        Return all records from call
        """

        header: dict[str, str] = SoliscloudAPI._prepare_header(key_id, secret,
            params, canonicalized_resource)

        url = f"{self.domain}{canonicalized_resource}"
        try:
            result = await self._post_data_json(url, header, params)
            return result['page']['records']
        except KeyError as err:
            raise SoliscloudAPI.ApiError("Malformed data", result) from err

    async def _get_data(self, canonicalized_resource: str, key_id: str, secret: bytes, params: dict[str, Any]):
        """
        Return data from call
        """

        header: dict[str, str] = SoliscloudAPI._prepare_header(key_id, secret,
            params, canonicalized_resource)

        url = f"{self.domain}{canonicalized_resource}"
        result = await self._post_data_json(url, header, params)

        return result

    @staticmethod
    def _now() -> datetime.datetime:
        return datetime.now(timezone.utc)

    @staticmethod
    def _prepare_header(
        key_id: str,
        secret: bytes,
        body: dict[str, str],
        canonicalized_resource: str
    ) -> dict[str, str]:
        content_md5 = base64.b64encode(
            hashlib.md5(json.dumps(body, separators=(",", ":")).encode('utf-8')).digest()
        ).decode('utf-8')

        content_type = "application/json"

        date = SoliscloudAPI._now().strftime("%a, %d %b %Y %H:%M:%S GMT")

        encrypt_str = (VERB + "\n"
            + content_md5 + "\n"
            + content_type + "\n"
            + date + "\n"
            + canonicalized_resource
        )
        hmac_obj = hmac.new(
            secret,
            msg=encrypt_str.encode('utf-8'),
            digestmod=hashlib.sha1
        )
        sign = base64.b64encode(hmac_obj.digest())
        authorization = "API " + key_id + ":" + sign.decode('utf-8')

        header: dict[str, str] = {
            "Content-MD5": content_md5,
            "Content-Type": content_type,
            "Date": date,
            "Authorization": authorization
        }
        return header

    async def _post_data_json(self,
        url: str,
        header: dict[str, Any],
        params: dict[str, Any]
    ) -> dict[str, Any]:
        """ Http-post data to specified domain/canonicalized_resource. """

        resp = None
        result = None
        if self._session is None:
            raise SoliscloudAPI.SolisCloudError("aiohttp.ClientSession not set")
        try:
            async with async_timeout.timeout(10):
                resp = await SoliscloudAPI._do_post_aiohttp(self._session, url, params, header)

                result = await resp.json()
                if resp.status == HTTPStatus.OK:
                    if result['code'] != '0':
                        raise SoliscloudAPI.ApiError(result['msg'], result['code'])
                    return result['data']
                else:
                    raise SoliscloudAPI.HttpError(resp.status)
        except asyncio.TimeoutError as err:
            if resp is not None:
                await resp.release()
            raise SoliscloudAPI.TimeoutError() from err
        except ClientError as err:
            if resp is not None:
                await resp.release()
            raise SoliscloudAPI.ApiError(err)
        except (KeyError, TypeError) as err:
            raise SoliscloudAPI.ApiError("Malformed server response",
                response=result) from err

    @staticmethod
    async def _do_post_aiohttp(
        session,
        url: str,
        params: dict[str, Any],
        header: dict[str, Any]
    ) -> dict[str, Any]:
        """ Allows mocking for unit tests."""
        return await session.post(url, json=params, headers=header)

    @staticmethod
    def _verify_date(format: SoliscloudAPI.DateFormat, date: str):
        rex = re.compile("^[0-9]{4}-[0-9]{2}-[0-9]{2}$")
        err = SoliscloudAPI.SolisCloudError("time must be in format YYYY-MM-DD")
        if format == SoliscloudAPI.DateFormat.MONTH:
            rex = re.compile("^[0-9]{4}-[0-9]{2}$")
            err = SoliscloudAPI.SolisCloudError("month must be in format YYYY-MM")
        elif format == SoliscloudAPI.DateFormat.YEAR:
            rex = re.compile("^[0-9]{4}$")
            err = SoliscloudAPI.SolisCloudError("year must be in format YYYY")
        if not rex.match(date):
            raise err
        return

class SolisData:
    def __init__(self, settings: SolisSettings):
        self.domain = f"{settings.api_url}:{settings.port}"
        self.api_key = settings.api_key
        api_secret_str = settings.client_secret
        if not self.api_key or not api_secret_str:
            raise ValueError("SOLIS_CLOUD_API_KEY or SOLIS_CLOUD_API_KEY_SECRET environment variable is not set")
        self.api_secret = api_secret_str.encode('utf-8')  # Convert to binary string

    async def get_inverter_list(self, soliscloud: SoliscloudAPI):
        """Fetch the list of inverters"""
        inverter_list = await soliscloud.inverter_list(
            self.api_key, 
            self.api_secret, 
            page_no=1, 
            page_size=100
        )
        return inverter_list
    
    def process_solis_data(self, live_generation_kw: pd.DataFrame) -> pd.DataFrame:
        """
        Process the Solis data and convert it to a DataFrame with timestamp and power_kw columns.
        
        :param live_generation_kw: DataFrame with original Solis data
        :return: DataFrame with processed data
        """
        # Create a copy of the DataFrame to avoid SettingWithCopyWarning
        processed_df = live_generation_kw[['timestamp', 'power_kw']].copy()
        
        # Ensure the timestamp is in the correct format
        processed_df.loc[:, 'timestamp'] = pd.to_datetime(processed_df['timestamp'])
        
        # Sort by timestamp
        processed_df = processed_df.sort_values('timestamp')
        
        # Reset the index
        processed_df = processed_df.reset_index(drop=True)
        
        return processed_df

    async def get_solis_data(self) -> pd.DataFrame:
        """
        Get live PV generation data from Solis API for the last 7 days
        :return: DataFrame with timestamp and power_kw columns
        """
        async with ClientSession() as websession:
            soliscloud = SoliscloudAPI(self.domain, websession)
            
            inverter_list = await self.get_inverter_list(soliscloud)
            if not inverter_list:
                raise ValueError("No inverters found")

            end_time = datetime.now(timezone.utc)
            start_time = end_time - timedelta(days=7)
            
            data_list = []
            
            for inverter in inverter_list:
                inverter_sn = inverter['sn']
                for day in range(7):
                    current_date = (end_time - timedelta(days=day)).strftime('%Y-%m-%d')
                    try:
                        inverter_day_data = await soliscloud.inverter_day(
                            self.api_key,
                            self.api_secret,
                            currency='USD',
                            time=current_date,
                            time_zone=0,
                            inverter_sn=inverter_sn
                        )
                        
                        # Check if inverter_day_data is a list of dictionaries
                        if isinstance(inverter_day_data, list) and all(isinstance(item, dict) for item in inverter_day_data):
                            for data_point in inverter_day_data:
                                timestamp = datetime.fromtimestamp(int(data_point['dataTimestamp']) / 1000, tz=timezone.utc)
                                if start_time <= timestamp <= end_time:
                                    data_list.append({
                                        "timestamp": timestamp.strftime('%Y-%m-%d %H:%M:%S'),
                                        "power_kw": float(data_point['pac']) / 1000,  # Convert W to kW
                                        "inverter_sn": inverter_sn
                                    })
                        else:
                            print(f"Unexpected data format for inverter {inverter_sn} on {current_date}")
                            print(f"Received data: {inverter_day_data}")
                    
                    except Exception as e:
                        print(f"Error fetching data for inverter {inverter_sn} on {current_date}: {e}")
                        print(f"Received data: {inverter_day_data}")
                    
                    # Avoid rate limiting
                    await asyncio.sleep(0.5)  # 2 times/sec limit
            
            # Convert the list to a DataFrame
            live_generation_kw = pd.DataFrame(data_list)
            
            if live_generation_kw.empty:
                return pd.DataFrame(columns=["timestamp", "power_kw"])

            # Convert to datetime
            live_generation_kw["timestamp"] = pd.to_datetime(live_generation_kw["timestamp"])
            
            # Sort by timestamp
            live_generation_kw = live_generation_kw.sort_values("timestamp")
            
            # Process the data to match the desired format
            processed_df = self.process_solis_data(live_generation_kw)
            processed_df = processed_df.reset_index(drop=True)
            
            return processed_df


async def get_solis_data(settings: SolisSettings):
    solis_data = SolisData(settings)
    return await solis_data.get_solis_data()



import pytest
import pandas as pd
import numpy as np
import xarray as xr
from datetime import datetime, timezone
from quartz_solar_forecast.data import process_pv_data  
from quartz_solar_forecast.pydantic_models import PVSite

@pytest.fixture
def sample_site():
    return PVSite(
        latitude=51.75,
        longitude=-1.25,
        capacity_kwp=1.25,
        tilt=35,
        orientation=180,
        inverter_type="enphase"
    )

@pytest.fixture
def sample_timestamp():
    timestamp = datetime.now().timestamp()
    timestamp_str = datetime.fromtimestamp(timestamp, tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
    return pd.to_datetime(timestamp_str)

@pytest.fixture
def sample_live_generation():
    return pd.DataFrame({
        'timestamp': [
            pd.Timestamp('2024-06-16 10:00:00'),
            pd.Timestamp('2024-06-16 10:05:00'),
            pd.Timestamp('2024-06-16 10:10:00')
        ],
        'power_kw': [0.75, 0.80, 0.78]
    })

def test_process_pv_data_with_live_data(sample_site, sample_timestamp, sample_live_generation):
    result = process_pv_data(sample_live_generation, sample_timestamp, sample_site)

    assert isinstance(result, xr.Dataset)
    assert 'generation_kw' in result.data_vars
    assert set(result.coords) == {'longitude', 'latitude', 'timestamp', 'pv_id', 'kwp', 'tilt', 'orientation'}
    assert result.pv_id.values.tolist() == [1]
    assert result.longitude.values.tolist() == [sample_site.longitude]
    assert result.latitude.values.tolist() == [sample_site.latitude]
    assert result.kwp.values.tolist() == [sample_site.capacity_kwp]
    assert result.tilt.values.tolist() == [sample_site.tilt]
    assert result.orientation.values.tolist() == [sample_site.orientation]
    assert len(result.timestamp) <= len(sample_live_generation)
    assert np.all(result.timestamp.values <= sample_timestamp)

def test_process_pv_data_without_live_data(sample_site, sample_timestamp):
    result = process_pv_data(None, sample_timestamp, sample_site)

    assert isinstance(result, xr.Dataset)
    assert 'generation_kw' in result.data_vars
    assert set(result.coords) == {'longitude', 'latitude', 'timestamp', 'pv_id', 'kwp', 'tilt', 'orientation'}
    assert result.pv_id.values.tolist() == [1]
    assert result.longitude.values.tolist() == [sample_site.longitude]
    assert result.latitude.values.tolist() == [sample_site.latitude]
    assert result.kwp.values.tolist() == [sample_site.capacity_kwp]
    assert result.tilt.values.tolist() == [sample_site.tilt]
    assert result.orientation.values.tolist() == [sample_site.orientation]
    assert len(result.timestamp) == 1
    assert result.timestamp.values[0] == sample_timestamp
    assert np.isnan(result.generation_kw.values[0][0])


""" Function to get NWP data and create fake PV dataset"""
import ssl
from datetime import datetime
from typing import Optional

import numpy as np
import openmeteo_requests
import pandas as pd
import requests_cache
import xarray as xr
from retry_requests import retry

from quartz_solar_forecast.pydantic_models import PVSite

ssl._create_default_https_context = ssl._create_unverified_context


def get_nwp(site: PVSite, ts: datetime, nwp_source: str = "icon") -> xr.Dataset:
    """
    Get GFS NWP data for a point time space and time

    :param site: the PV site
    :param ts: the timestamp for when you want the forecast for
    :param nwp_source: the nwp data source. Either "gfs", "icon" or "ukmo". Defaults to "icon"
    :return: nwp forecast in xarray
    """

    # Setup the Open-Meteo API client with cache and retry on error
    cache_session = requests_cache.CachedSession('.cache', expire_after = -1)
    retry_session = retry(cache_session, retries = 5, backoff_factor = 0.2)
    openmeteo = openmeteo_requests.Client(session = retry_session)

    # Define the variables we want. Visibility is handled separately after the main request
    variables = [
        "temperature_2m",
        "precipitation",
        "cloud_cover_low",
        "cloud_cover_mid",
        "cloud_cover_high",
        "wind_speed_10m",
        "shortwave_radiation",
        "direct_radiation"
    ]

    start = ts.date()
    end = start + pd.Timedelta(days=7)

    url = ""

    # check whether the time stamp is more than 3 months in the past
    if (datetime.now() - ts).days > 90:
        print("Warning: The requested timestamp is more than 3 months in the past. The weather data are provided by a reanalyse model and not ICON or GFS.")

        # load data from open-meteo Historical Weather API
        url = "https://archive-api.open-meteo.com/v1/archive"

    else:
        # Getting NWP from open meteo weather forecast API by ICON, GFS, or UKMO within the last 3 months
        if nwp_source == "icon":
            url_nwp_source = "dwd-icon"
            url = f"https://api.open-meteo.com/v1/{url_nwp_source}"
        elif nwp_source == "gfs":
            url_nwp_source = "gfs"
            url = f"https://api.open-meteo.com/v1/{url_nwp_source}"
        elif nwp_source == "ukmo":
            url = "https://api.open-meteo.com/v1/forecast"
        else:
            raise Exception(f'Source ({nwp_source}) must be either "icon", "gfs", or "ukmo"')

    params = {
        "latitude": site.latitude,
        "longitude": site.longitude,
        "start_date": f"{start}",
        "end_date": f"{end}",
        "hourly": variables
    }

    # Add the "models" parameter if using "ukmo"
    if nwp_source == "ukmo":
        params["models"] = "ukmo_seamless"

    # Make API call to URL
    response = openmeteo.weather_api(url, params=params)
    hourly = response[0].Hourly()

    hourly_data = {"time": pd.date_range(
    	start = pd.to_datetime(hourly.Time(), unit = "s", utc = False),
    	end = pd.to_datetime(hourly.TimeEnd(), unit = "s", utc = False),
    	freq = pd.Timedelta(seconds = hourly.Interval()),
    	inclusive = "left"
    )}


    # variables index as in the variables array of the request
    hourly_data["t"] = hourly.Variables(0).ValuesAsNumpy()
    hourly_data["prate"] = hourly.Variables(1).ValuesAsNumpy()
    hourly_data["lcc"] = hourly.Variables(2).ValuesAsNumpy()
    hourly_data["mcc"] = hourly.Variables(3).ValuesAsNumpy()
    hourly_data["hcc"] = hourly.Variables(4).ValuesAsNumpy()
    hourly_data["si10"] = hourly.Variables(5).ValuesAsNumpy()
    hourly_data["dswrf"] = hourly.Variables(6).ValuesAsNumpy()
    hourly_data["dlwrf"] = hourly.Variables(7).ValuesAsNumpy()

    # handle visibility
    if (datetime.now() - ts).days <= 90:
        # load data from open-meteo gfs model
        params = {
        	"latitude": site.latitude,
        	"longitude": site.longitude,
        	"start_date": f"{start}",
        	"end_date": f"{end}",
        	"hourly": "visibility"
        }
        data_vis_gfs = openmeteo.weather_api("https://api.open-meteo.com/v1/gfs", params=params)[0].Hourly().Variables(0).ValuesAsNumpy()
        hourly_data["vis"] = data_vis_gfs
    else:
        # set to maximum visibility possible
        hourly_data["vis"] = 24000.0

    df = pd.DataFrame(data = hourly_data)
    df = df.set_index("time")
    df = df.astype('float64')

    # convert data into xarray
    data_xr = format_nwp_data(df, nwp_source, site)

    return data_xr

def format_nwp_data(df: pd.DataFrame, nwp_source:str, site: PVSite):
    data_xr = xr.DataArray(
        data=df.values,
        dims=["step", "variable"],
        coords=dict(
            step=("step", df.index - df.index[0]),
            variable=df.columns,
        ),
    )
    data_xr = data_xr.to_dataset(name=nwp_source)
    data_xr = data_xr.assign_coords(
        {"x": [site.longitude], "y": [site.latitude], "time": [df.index[0]]}
    )
    return data_xr


def process_pv_data(live_generation_kw: Optional[pd.DataFrame], ts: pd.Timestamp, site: 'PVSite') -> xr.Dataset:
    """
    Process PV data and create an xarray Dataset.

    :param live_generation_kw: DataFrame containing live generation data, or None
    :param ts: Current timestamp
    :param site: PV site information
    :return: xarray Dataset containing processed PV data
    """
    if live_generation_kw is not None and not live_generation_kw.empty:
        # Get the most recent data
        recent_pv_data = live_generation_kw[live_generation_kw['timestamp'] <= ts]
        power_kw = np.array([recent_pv_data["power_kw"].values], dtype=np.float64)
        timestamp = recent_pv_data['timestamp'].values
    else:
        # Make fake PV data; this is where we could add the history of a PV system
        power_kw = np.array([[np.nan]])
        timestamp = [ts]

    da = xr.DataArray(
        data=power_kw,
        dims=["pv_id", "timestamp"],
        coords=dict(
            longitude=(["pv_id"], [site.longitude]),
            latitude=(["pv_id"], [site.latitude]),
            timestamp=timestamp,
            pv_id=[1],
            kwp=(["pv_id"], [site.capacity_kwp]),
            tilt=(["pv_id"], [site.tilt]),
            orientation=(["pv_id"], [site.orientation]),
        ),
    )
    da = da.to_dataset(name="generation_kw")

    return da

def make_pv_data(site: PVSite, ts: pd.Timestamp) -> xr.Dataset:
    """
    Make PV data by combining live data from various inverters.
    
    :param site: the PV site
    :param ts: the timestamp of the site
    :return: The combined PV dataset in xarray form
    """
    live_generation_kw = site.get_inverter().get_data(ts)
    # Process the PV data
    da = process_pv_data(live_generation_kw, ts, site)

    return da



import os
from datetime import datetime, timezone
from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
import pandas as pd
from dotenv import load_dotenv
from quartz_solar_forecast.forecast import run_forecast
from quartz_solar_forecast.pydantic_models import PVSite, ForecastRequest, TokenRequest
from quartz_solar_forecast.inverters.enphase import get_enphase_auth_url, get_enphase_access_token

load_dotenv()

app = FastAPI()

# CORS middleware setup
origins = [
    "http://localhost:5173",
    "localhost:5173",
    "http://localhost:8501",
    "localhost:8501"
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"]
)

@app.post("/forecast/")
def forecast(forecast_request: ForecastRequest):
    site = forecast_request.site
    ts = forecast_request.timestamp if forecast_request.timestamp else datetime.now(timezone.utc).isoformat()

    timestamp = pd.Timestamp(ts).tz_localize(None)
    formatted_timestamp = timestamp.strftime('%Y-%m-%d %H:%M:%S')

    site_no_live = PVSite(latitude=site.latitude, longitude=site.longitude, capacity_kwp=site.capacity_kwp)
    predictions_no_live = run_forecast(site=site_no_live, ts=timestamp)

    if not site.inverter_type:
        predictions = predictions_no_live
    else:
        predictions_with_live = run_forecast(site=site, ts=timestamp)
        predictions_with_live['power_kw_no_live_pv'] = predictions_no_live['power_kw']
        predictions = predictions_with_live

    response = {
        "timestamp": formatted_timestamp,
        "predictions": predictions.to_dict(),
    }

    return response

@app.get("/solar_inverters/enphase/auth_url")
def get_enphase_authorization_url():
    auth_url = get_enphase_auth_url()
    return {"auth_url": auth_url}

@app.post("/solar_inverters/enphase/token_and_id")
def get_enphase_token_and_system_id(request: TokenRequest):
    if "?code=" not in request.redirect_url:
        raise HTTPException(status_code=400, detail="Invalid redirect URL")
    
    auth_code = request.redirect_url.split("?code=")[1]
    try:
        access_token = get_enphase_access_token(auth_code)
        enphase_system_id = os.getenv("ENPHASE_SYSTEM_ID")
        return {"access_token": access_token, "enphase_system_id": enphase_system_id}
    except Exception as e:
        raise HTTPException(status_code=400, detail=f"Error getting access token and system ID: {str(e)}")


from quartz_solar_forecast.evaluation import run_eval
import tempfile
import pandas as pd


def test_run_eval():

    # create a fake dataframe

    with tempfile.TemporaryDirectory() as tmpdirname:
        print("created temporary directory", tmpdirname)

        test_dataset = pd.DataFrame(
            columns=[
                "pv_id",
                "timestamp",
            ],
            data=[[8215, "2021-01-26 01:15:00"], [8215, "2021-01-30 16:30:00"]],
        )

        testset_filename = tmpdirname + "/test_dataset.csv"
        test_dataset.to_csv(testset_filename, index=False)

        # call the metrics function
        run_eval(testset_filename)



from quartz_solar_forecast.eval.pv import get_pv_truth, get_pv_metadata
import pandas as pd


def test_get_pv_metadata():
    test_set_df = pd.DataFrame(
        [
            {
                "timestamp": pd.Timestamp("2021-01-26 01:15:00"),
                "pv_id": 8215,
            }
        ]
    )

    metadata_df = get_pv_metadata(test_set_df)
    assert "latitude" in metadata_df.columns


def test_get_pv():
    # make test dataset file
    test_set_df = pd.DataFrame(
        [
            {
                "timestamp": pd.Timestamp("2021-01-26 01:15:00"),
                "pv_id": 8215,
            }
        ]
    )

    # Collect NWP data from Hugging Face, ICON. (Peter)
    _ = get_pv_truth(test_set_df)



"""
Make a random test set

This takes a random subset of times and for various pv ids and makes a test set

There is an option to odmit timestamps that don't exsits in the ICON dataset:
https://huggingface.co/datasets/openclimatefix/dwd-icon-eu/tree/main/data
"""
import os
from typing import Optional

import numpy as np
import pandas as pd

from quartz_solar_forecast.eval.utils import make_hf_filename
from huggingface_hub import HfFileSystem

test_start_date = pd.Timestamp("2021-01-01")
test_end_date = pd.Timestamp("2022-01-01")

# this have been chosen from the entire training set. This ideas
pv_ids = [
    9531,
    7174,
    6872,
    7386,
    13607,
    6330,
    26841,
    6665,
    4045,
    26846,
    6494,
    7834,
    3543,
    7093,
    3864,
    8412,
    3454,
    9765,
    10585,
    26942,
    7721,
    26804,
    7551,
    26861,
    7568,
    7338,
    7410,
    6967,
    16480,
    7241,
    7593,
    7557,
    7757,
    3094,
    6800,
    26905,
    5512,
    26840,
    7595,
    5803,
    26876,
    7846,
    26786,
    7580,
    6629,
    16477,
    3489,
    26796,
    12761,
    26903,
]

np.random.seed(42)


def make_test_set(output_file_name: Optional[str] = None, number_of_samples_per_system: int = 50, check_hf_files: bool = False):
    """
    Make a test set of random times and pv ids

    :param output_file_name: the name of the file to write the test set to
    :param number_of_samples_per_system: the number of samples to take per pv id
    """

    if output_file_name is None:
        # get the folder where this file is
        output_file_name = os.path.dirname(os.path.abspath(__file__)) + "/testset.csv"

    ts = pd.date_range(start=test_start_date, end=test_end_date, freq="15min")

    # check that the files are in HF for ICON
    if check_hf_files:
        ts = filter_timestamps_if_hf_files_exists(ts)

    test_set = []
    for pv_id in pv_ids:
        ts_choice = ts[np.random.choice(len(ts), size=number_of_samples_per_system, replace=False)]
        test_set.append(pd.DataFrame({"pv_id": pv_id, "timestamp": ts_choice}))
    test_set = pd.concat(test_set)
    test_set.to_csv(output_file_name, index=False)

    return test_set


def filter_timestamps_if_hf_files_exists(timestamps_full: pd.DatetimeIndex):
    """
    Filter the timestamps if the huggingface files exist

    We are checking if the teimstamps, rounded down to the nearest 6 hours,
    exist in
    https://huggingface.co/datasets/openclimatefix/dwd-icon-eu/tree/main/data

    """
    timestamps = []
    fs = HfFileSystem()
    # print(fs.ls("datasets/openclimatefix/dwd-icon-eu/data/2022/4/11/", detail=False))
    for timestamp in timestamps_full:
        timestamp_floor = timestamp.floor("6h")
        _, huggingface_file = make_hf_filename(timestamp_floor)
        huggingface_file = huggingface_file[14:]

        if fs.exists(huggingface_file):
            timestamps.append(timestamp)
        else:
            print(f"Skipping {timestamp} because {huggingface_file} does not exist")

    timestamps = pd.DatetimeIndex(timestamps)
    return timestamps


# To run the script, un comment the following line and run this file
# make_test_set(check_hf_files=True)



import ocf_datapipes  # noqa
from ocf_datapipes.training.example.simple_pv import simple_pv_datapipe
from ocf_datapipes.training.example.pv_nwp import pv_nwp_datapipe
from ocf_datapipes.batch import BatchKey
from datetime import datetime
import os
import certifi
import ssl
import warnings

# Suppress warnings and configure SSL certificate for secure connections
warnings.filterwarnings("ignore")
os.environ['SSL_CERT_FILE'] = certifi.where()
ssl._create_default_https_context = ssl._create_unverified_context

def process_ocf_datapipes(config_file):

    """
    This function demonstrates the use of Open Climate Fix's datapipes with the Open Source Quartz Solar Forecast Project.
    It is an exploratory integration and not currently utilized in the Quartz Solar Forecast production pipeline.

    It processes solar power (PV) and numerical weather prediction (NWP) data to prepare it for forecasting tasks.

    :param config_file: The config file that specifies the paths to the data files, data preprocessing parameters, and other configurations necessary for the datapipes to function correctly.
    -  'pv_and_nwp_config.yaml' specifes the paths to necessary data files (e.g., PV output data, NWP data).
    -  The YAML file needs to be edited to reflect the correct paths to your data files and any specific preprocessing requirements for your project.

    This script is meant as a starting point for integrating Open Climate Fix datapipes into the Quartz Solar Forecast Project, serving as an example of how to preprocess and load data for solar power forecasting.
    """

    pv_data_pipe = simple_pv_datapipe(configuration_filename=config_file)
    pv_nwp_data_pipe = pv_nwp_datapipe(configuration_filename=config_file)

    ## Pv datapipe: store the first batch of processed data for inspection
    pv_batch = next(iter(pv_data_pipe))

    # Access the batch elements through batch keys: Convert the first set of timestamps to readable dates
    pv_times_readable = [datetime.utcfromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S') for ts in pv_batch[BatchKey.pv_time_utc][0]]
    print("Readable PV Times:", pv_times_readable)

    # Check Observed Capacity (Wp), and Nominal Capacity (Wp) of the PV system
    observed_capacity = pv_batch[BatchKey.pv_observed_capacity_wp]
    nominal_capacity = pv_batch[BatchKey.pv_nominal_capacity_wp]

    print("Observed Capacity (Wp):", observed_capacity)
    print("Nominal Capacity (Wp):", nominal_capacity)

    # NWP_PV DataPipe: Retrieve and print NWP data from the batch
    pv_nwp_batch = next(iter(pv_nwp_data_pipe))
    nwp_data = pv_nwp_batch[BatchKey.nwp]
    print("NWP data", nwp_data)


# Load configuration from YAML specifying data sources and processing parameters for datapipes
config_file = 'pv_and_nwp_config.yaml'
process_ocf_datapipes(config_file)


from quartz_solar_forecast.eval.nwp import get_nwp
import pandas as pd


# can take ~ 1 minute to run
def test_get_nwp():
    # make test dataset file
    test_set_df = pd.DataFrame(
        [
            {
                "timestamp": pd.Timestamp("2021-01-26 01:15:00"),
                "latitude": 51.5,
                "longitude": 0.0,
                "pv_id": 0,
            }
        ]
    )

    # Collect NWP data from Hugging Face, ICON. (Peter)
    _ = get_nwp(test_set_df)



from datetime import datetime


def get_file_path(latitude: float,
                  longitude: float,
                  capacity_kwp: float,
                  model: str = "gb",
                  time: datetime = None) -> str:
    return time.strftime(f"data/%Y/%-m/%-d/{model}_{latitude}_{longitude}_{capacity_kwp}_%Y%m%d_%H.csv")



if __name__ == "__main__":
    # please change the site name, start_datetime and end_datetime, latitude, longitude and capacity_kwp as per your requirement
    write_out_forecasts(
        init_time_freq=6,
        start_datetime="2024-03-10 00:00:00",
        end_datetime="2024-03-11 00:00:00",
        site_name="Test",
        latitude=51.75,
        longitude=-1.25,
        capacity_kwp=1.25
    )


import pytest
import pandas as pd
import numpy as np
from quartz_solar_forecast.inverters.enphase import process_enphase_data 

@pytest.fixture
def sample_data():
    return {
        'system_id': 3136663,
        'granularity': 'week',
        'total_devices': 4,
        'start_at': 1718530896,
        'end_at': 1719134971,
        'items': 'intervals',
        'intervals': [
            {'end_at': 1718531100, 'devices_reporting': 4, 'powr': 624, 'enwh': 52},
            {'end_at': 1718531400, 'devices_reporting': 4, 'powr': 684, 'enwh': 57},
            {'end_at': 1718531700, 'devices_reporting': 4, 'powr': 672, 'enwh': 56},
        ]
    }

def test_process_enphase_data(sample_data):
    # Set start_at to before/after the first interval
    start_at = sample_data['intervals'][0]['end_at'] + 1 
    
    # Process the data
    result = process_enphase_data(sample_data, start_at)
    
    # Check if the result is a DataFrame
    assert isinstance(result, pd.DataFrame)
    
    # Check if the DataFrame has the expected columns
    assert set(result.columns) == {'timestamp', 'power_kw'}
    
    # Check if the timestamp column is of datetime type
    assert pd.api.types.is_datetime64_any_dtype(result['timestamp'])
    
    # Check if power_kw values are correctly calculated (divided by 1000)
    expected_power_values = [interval['powr'] / 1000 for interval in sample_data['intervals']]
    assert all(value in expected_power_values for value in result['power_kw'])
    
    # Convert start_at to a naive UTC timestamp
    start_at_timestamp = pd.Timestamp(start_at, unit='s').tz_localize('UTC').tz_convert(None)
    
    # Check if all timestamps are after the start_at time
    assert np.all(result['timestamp'] >= start_at_timestamp)
    
    # Check if the number of rows is less than or equal to the number of intervals
    assert len(result) <= len(sample_data['intervals'])

    # Check if timestamps are formatted correctly
    expected_timestamps = [
        pd.Timestamp(interval['end_at'], unit='s').tz_localize('UTC').tz_convert(None).strftime('%Y-%m-%d %H:%M:%S')
        for interval in sample_data['intervals']
    ]
    assert all(ts.strftime('%Y-%m-%d %H:%M:%S') in expected_timestamps for ts in result['timestamp'])


from datetime import datetime, timedelta
import logging

import pandas as pd

from quartz_solar_forecast.data import get_nwp, make_pv_data
from quartz_solar_forecast.forecasts import forecast_v1_tilt_orientation, TryolabsSolarPowerPredictor
from quartz_solar_forecast.pydantic_models import PVSite

log = logging.getLogger(__name__)

def predict_ocf(
    site: PVSite, model=None, ts: datetime | str = None, nwp_source: str = "icon"
):
    """
    Run the forecast with the gb model, which can take tilt and orientation as inputs
    
    :param site: the PV site
    :param model: the model to use for prediction
    :param ts: the timestamp of the site. If None, defaults to the current timestamp rounded down to 15 minutes.
    :param nwp_source: the nwp data source. Either "gfs", "icon" or "ukmo". Defaults to "icon" 
    :return: The PV forecast of the site for time (ts) for 48 hours
    """
    if ts is None:
        ts = pd.Timestamp.now().round("15min")

    if isinstance(ts, str):
        ts = datetime.fromisoformat(ts)

    if site.capacity_kwp > 4:
        log.warning("Your site capacity is greater than 4kWp, "
                    "however the model is trained on sites with capacity <= 4kWp."
                    "We therefore will run the model with a capacity of 4 kWp, "
                    "and we'll scale the results afterwards.")
        capacity_kwp_original = site.capacity_kwp
        site.capacity_kwp = 4
    else:
        capacity_kwp_original = site.capacity_kwp

    # make pv and nwp data from nwp_source
    nwp_xr = get_nwp(site=site, ts=ts, nwp_source=nwp_source)
    pv_xr = make_pv_data(site=site, ts=ts)

    # load and run models
    pred_df = forecast_v1_tilt_orientation(nwp_source, nwp_xr, pv_xr, ts, model=model)

    # scale the results if the capacity is different
    if capacity_kwp_original != site.capacity_kwp:
        pred_df["power_kw"] = pred_df["power_kw"] * capacity_kwp_original / site.capacity_kwp

    return pred_df


def predict_tryolabs(
    site: PVSite, ts: datetime | str = None):
    """
    Run the forecast with the xgb model
    
    :param site: the PV site
    :param ts: the timestamp of the site. If None, defaults to the current timestamp rounded down to 15 minutes.
    :return: The PV forecast of the site for time (ts) for 48 hours
    """

    # instantiate class to make predictions
    solar_power_predictor = TryolabsSolarPowerPredictor()
    
    # set start and end time, if no time is given use current time
    if ts is None:
        start_date = pd.Timestamp.now().strftime("%Y-%m-%d")
        start_time = pd.Timestamp.now().round(freq='h')
    else:
        start_date = pd.Timestamp(ts).strftime("%Y-%m-%d")
        start_time = pd.Timestamp(ts).round(freq='h')

    end_time = start_time + pd.Timedelta(hours=48)
    start_date_datetime = datetime.strptime(start_date, "%Y-%m-%d")

    # Check if the start date is more than 3 months ago
    three_months_ago = datetime.today() - timedelta(days=3 * 30)

    if start_date_datetime < three_months_ago:
        print(
            f"Start date ({start_date}) is more than 3 months ago, no",
            "forecast data available.",
        )
    else:
        # download the model from google drive and decompress if necessary
        solar_power_predictor.load_model()
        # make predictions
        predictions = solar_power_predictor.predict_power_output(
            latitude=site.latitude,
            longitude=site.longitude,
            start_date=start_date,
            kwp=site.capacity_kwp,
            orientation=site.orientation,
            tilt=site.tilt,
        )

        # postprocessing of the dataframe
        predictions = predictions[
            (predictions["date"] >= start_time) & (predictions["date"] < end_time)
        ]
        predictions = predictions.reset_index(drop=True)
        predictions.set_index("date", inplace=True)
        print("Predictions finished.")
        return predictions


def run_forecast(
    site: PVSite,
    model: str = "gb",
    ts: datetime | str = None,
    nwp_source: str = "icon",
) -> pd.DataFrame:
    """
    Predict solar power output for a given site using a specified model.

    :param site: the PV site
    :param model: the model to use for prediction, choose between "ocf" and "tryolabs",
                    by default "ocf" is used
    :param ts: the timestamp of the site. If None, defaults to the current timestamp rounded down to 15 minutes.
    :param nwp_source: the nwp data source. Either "gfs", "icon" or "ukmo". Defaults to "icon" 
                       (only relevant if model=="gb")
    :return: The PV forecast of the site for time (ts) for 48 hours
    """

    if model == "gb":
        return predict_ocf(site, None, ts, nwp_source)
              
    elif model == "xgb":
        return predict_tryolabs(site, ts)
    
    else:  
        raise ValueError(f"Unsupported model: {model}. Choose between 'xgb' and 'gb'")



""" This script is used to run the forecast every day at 9am

Github actions runs this script every day at 9am to generate the forecast for the next 48 hours.
The results are saved to HF -
https://huggingface.co/openclimatefix/open-source-quartz-solar-forecast/tree/main/data

This means we can start to compare the difference between the forecasts and the actual generation.
"""

import os
from datetime import datetime
from huggingface_hub import login, HfFileSystem, HfApi
from quartz_solar_forecast.utils.forecast_csv import forecast_for_site
from quartz_solar_forecast.utils.file_path import get_file_path


if __name__ == "__main__":

    hf_token = os.getenv("HF_TOKEN")
    hf_repo = os.getenv("HF_REPO")
    print(hf_repo)

    login(hf_token)
    fs = HfFileSystem(token=hf_token)
    now = datetime.utcnow()
    latitude = 51.59
    longitude = -1.89
    capacity_kwp = 4

    for model in ["gb", "xgb"]:
        forecast = forecast_for_site(latitude, longitude, capacity_kwp, model, now)

        path = get_file_path(latitude, longitude, capacity_kwp, model, now)
        with fs.open(f"{hf_repo}/{path}", "w") as f:
            forecast.to_csv(path_or_buf=f)



import datetime
import pandas as pd
import zipfile
import os.path
import shutil
import logging

from huggingface_hub import hf_hub_download
from quartz_solar_forecast.weather import WeatherService

from xgboost.sklearn import XGBRegressor

from . import constants
import quartz_solar_forecast

logger = logging.getLogger(__name__)

class TryolabsSolarPowerPredictor:
    """
    A class to predict solar power output based on weather data, location, panel orientation,
        and other factors.

    Attributes
    ----------
    model_path : str
        Path to the trained model joblib file.

    Methods
    -------
    load_model -> None:

        Downloads the model from Google Drive and decropresses it if necessary
    
    predict_power_output(latitude: float, longitude: float, start_date: str, kwp: float,
        orientation: float, tilt: float) -> pd.DataFrame:

        Predicts solar power output for the given parameters.
    """
    DATE_COLUMN = "date"
    download_dir = os.path.dirname(quartz_solar_forecast.__file__) + "/models"

    def _download_model(self, filename: str, repo_id: str, file_path: str) -> str:
        """
        Downloads a model file from a Hugging Face repository and saves it locally.
    
        Parameters:
        -----------
        filename : str
            The name to give the downloaded file when saving it locally.
        repo_id : str
            The ID of the Hugging Face repository containing the model.
        file_path : str
            The path to the model file within the repository.
    
        Returns:
        --------
        str
            The path to the locally saved model file.
        """
        # Use the project directory instead of the user's home directory
        os.makedirs(self.download_dir, exist_ok=True)
        
        downloaded_file = hf_hub_download(repo_id=repo_id, filename=file_path, cache_dir=self.download_dir)
        
        target_path = os.path.join(self.download_dir, filename)

        # copy file from downloaded_file to target_path
        shutil.copyfile(downloaded_file, target_path)
        
        return target_path

    def _decompress_zipfile(self, filename: str) -> None:
        """
        Extract all files contained in a .zip file to the current directory.
        filename must contain .zip extension

        Parameters
        ----------
        filename : str
            The name of the .zip file to be decompressed
        """
        # get the directory of the file
        directory = os.path.dirname(filename)

        with zipfile.ZipFile(filename, "r") as zip_file:
            zip_file.extractall(path=directory)

    def load_model(
        self, 
        model_file: str = constants.MODEL_FILE,
        repo_id: str = "openclimatefix/open-source-quartz-solar-forecast",
        file_path: str = "models/v2/model_10_202405.ubj.zip"
    ) -> XGBRegressor:
        """
        Downloads, prepares, and loads the XGBoost model for solar power prediction.
    
        Parameters:
        -----------
        model_file : str, optional
            The name of the model file (without .zip extension).
            Default is set by constants.MODEL_FILE.
        repo_id : str, optional
            The ID of the Hugging Face repository containing the model.
            Default is "openclimatefix/open-source-quartz-solar-forecast".
        file_path : str, optional
            The path to the model file within the repository.
            Default is "models/v2/model_10_202405.ubj.zip".
    
        Returns:
        --------
        XGBRegressor
            The loaded XGBoost model ready for making predictions.
        """
        # Use the project directory
        zipfile_model = os.path.join(self.download_dir, model_file + ".zip")
    
        if not os.path.isfile(zipfile_model):
            logger.info("Downloading model...")
            zipfile_model = self._download_model(model_file + ".zip", repo_id, file_path)
        
        model_path = os.path.join(self.download_dir, model_file)
        if not os.path.isfile(model_path):
            logger.info("Preparing model...")
            self._decompress_zipfile(zipfile_model)

        logger.info("Loading model...")
        loaded_model = XGBRegressor()
        loaded_model.load_model(model_path)
        self.model = loaded_model
        return loaded_model
        
    def get_data(
        self,
        latitude: float,
        longitude: float,
        start_date: str,
        kwp: float,
        orientation: float = 180,
        tilt: float = 30,
    ) -> pd.DataFrame:
        """
        Fetches weather data for the given location and date range, and prepares it for prediction.

        Parameters
        ----------
        latitude : float
            Latitude of the location.
        longitude : float
            Longitude of the location.
        start_date : str
            Start date in 'YYYY-MM-DD' format.
        kwp : float
            Kilowatt peak of the solar panel system.
        orientation : float
            Orientation angle of the solar panel system in degrees.
        tilt : float
            Tilt angle of the solar panel system in degrees.

        Returns
        -------
        pd.DataFrame
            Prepared weather data with additional solar panel parameters.
        """
        start_date_datetime = datetime.datetime.strptime(start_date, "%Y-%m-%d")
        end_date_datetime = start_date_datetime + datetime.timedelta(days=2)
        end_date = end_date_datetime.strftime("%Y-%m-%d")

        weather_service = WeatherService()

        weather_data = weather_service.get_hourly_weather(
            latitude, longitude, start_date, end_date
        )

        PANEL_COLUMNS = [
            "latitude_rounded",
            "longitude_rounded",
            "orientation",
            "tilt",
            "kwp",
        ]

        weather_data["latitude_rounded"] = latitude
        weather_data["longitude_rounded"] = longitude
        weather_data["orientation"] = orientation
        weather_data["tilt"] = tilt
        weather_data["kwp"] = kwp

        cols = PANEL_COLUMNS + [
            col for col in weather_data.columns if col not in PANEL_COLUMNS
        ]
        weather_data = weather_data[cols]

        return weather_data

    def clean(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Cleans and transforms the input DataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            DataFrame containing the weather and solar panel data.

        Returns
        -------
        pd.DataFrame
            Transformed DataFrame ready for prediction.
        """
        df.loc[:, self.DATE_COLUMN] = pd.to_datetime(df[self.DATE_COLUMN])

        df.loc[:, "date_year"] = df[self.DATE_COLUMN].dt.year
        df.loc[:, "date_month"] = df[self.DATE_COLUMN].dt.month
        df.loc[:, "date_day"] = df[self.DATE_COLUMN].dt.day
        df.loc[:, "date_hour"] = df[self.DATE_COLUMN].dt.hour
        df.loc[:, "date_minute"] = df[self.DATE_COLUMN].dt.minute

        COLUMNS_TO_DROP = ["date_minute", "date_year",
                           "terrestrial_radiation",
                           "shortwave_radiation",
                           "direct_normal_irradiance"]

        df = df.drop(columns=COLUMNS_TO_DROP)

        return df

    def predict_power_output(
        self,
        latitude: float,
        longitude: float,
        start_date: str,
        kwp: float,
        orientation: float = 180,
        tilt: float = 30,
    ) -> pd.DataFrame:
        """
        Predicts solar power output for the specified parameters.

        Parameters
        ----------
        latitude : float
            Latitude of the location.
        longitude : float
            Longitude of the location.
        start_date : str
            Start date in 'YYYY-MM-DD' format.
        kwp : float
            Kilowatt peak of the solar panel system.
        orientation : float
            Orientation angle of the solar panel system in degrees.
        tilt : float
            Tilt angle of the solar panel system in degrees.

        Returns
        -------
        pd.DataFrame
            DataFrame containing timestamps and predicted power output in kW for every 15 minutes.
        """

        data = self.get_data(latitude, longitude, start_date, kwp, orientation, tilt)
        #if data is not None:
        cleaned_data = self.clean(data)
        predictions = self.model.predict(cleaned_data.drop(columns=[self.DATE_COLUMN]))
        predictions_df = pd.DataFrame(predictions, columns=["prediction"])
        final_data = cleaned_data.join(predictions_df)
        # set night predictions to 0
        final_data.loc[final_data["is_day"] == 0, "prediction"] = 0
        # set negative output to 0
        final_data.loc[final_data["prediction"] < 0, "prediction"] = 0
        df = final_data[[self.DATE_COLUMN, "prediction"]]
        df = df.rename(columns={"prediction": "power_kw"})
        return df



import os
import pandas as pd
from datetime import datetime, timedelta
import quartz_solar_forecast.forecast as forecast
from quartz_solar_forecast.utils.forecast_csv import write_out_forecasts
from quartz_solar_forecast.pydantic_models import PVSite

def test_generate_forecast(monkeypatch):
    site_name = "TestCase"
    latitude = 51.75
    longitude = -1.25
    capacity_kwp = 1.25
    start_datetime = "2024-03-10 00:00:00"
    end_datetime = "2024-03-11 00:00:00"
    init_time_freq = 6
    output_dir = os.path.join(os.getcwd(), "csv_forecasts")
    output_file_name = (
        f"forecast_{site_name}_{start_datetime[:10]}_{end_datetime[:10]}.csv"
    )
    output_file_path = os.path.join(output_dir, output_file_name)

    def mock_forecast(
        site: PVSite,
        model: str = "gb",
        ts: datetime | str = None,
        nwp_source: str = "icon",
    ):
        return pd.DataFrame(
            {
                "datetime": [
                    datetime(2024, 3, 10, 0, 0) + timedelta(hours=6 * i)
                    for i in range(4)
                ],
                "power_kw": [0.1, 0.5, 0.8, 0.6],
                "forecast_init_time": [datetime(2024, 3, 10, 0, 0)] * 4,
            }
        )

    monkeypatch.setattr(forecast, "run_forecast", mock_forecast)

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    write_out_forecasts(
        init_time_freq,
        start_datetime,
        end_datetime,
        site_name,
        latitude,
        longitude,
        capacity_kwp,
    )

    assert os.path.exists(output_file_path)
    os.remove(output_file_path)



from quartz_solar_forecast.eval.forecast import run_forecast

import pandas as pd


def test_run_forecast():
    pv_df = pd.DataFrame(
        [
            {
                "timestamp": pd.Timestamp("2021-01-26 01:15:00"),
                "latitude": 51.5,
                "longitude": 0.0,
                "capacity": 1.0,
                "pv_id": 1.0,
            }
        ]
    )

    nwp_df = pd.DataFrame(
        columns=[
            "pv_id",
            "timestamp",
            "t",
            "prate",
            "dswrf",
            "dlwrf",
            "lcc",
            "mcc",
            "hcc",
            "vis",
            "si10",
            "time",
        ],
        data=[
            [
                1,
                pd.Timestamp("2021-01-26 01:15:00"),
                1.0,
                1.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0,
                0,
                pd.Timestamp("2021-01-26 01:15:00"),
            ],
            [
                1,
                pd.Timestamp("2021-01-26 01:15:00"),
                1.0,
                1.0,
                1.0,
                1.0,
                0.0,
                0.0,
                0.0,
                0,
                0,
                pd.Timestamp("2021-01-26 01:30:00"),
            ],
        ],
    )

    # format timestamp
    pv_df["timestamp"] = pd.to_datetime(pv_df["timestamp"])
    nwp_df["timestamp"] = pd.to_datetime(nwp_df["timestamp"])
    nwp_df["time"] = pd.to_datetime(nwp_df["time"])

    _ = run_forecast(pv_df=pv_df, nwp_df=nwp_df, nwp_source="ICON")



from typing import Optional

import requests
import pandas as pd
from datetime import datetime

from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict

from quartz_solar_forecast.inverters.inverter import AbstractInverter


class GivEnergySettings(BaseSettings):
    model_config = SettingsConfigDict(env_file='.env', extra='ignore')

    api_key: str = Field(alias="GIVENERGY_API_KEY")


class GivEnergyInverter(AbstractInverter):

    def __init__(self, settings: GivEnergySettings):
        self.__settings = settings

    def get_data(self, ts: pd.Timestamp) -> Optional[pd.DataFrame]:
        try:
            return get_givenergy_data(self.__settings)
        except Exception as e:
            print(f"Error retrieving GivEnergy data: {e}")
            return None


def get_inverter_serial_number(settings: GivEnergySettings):
    """
    Fetch the inverter serial number from the GivEnergy communication device API.
    
    :return: Inverter serial number as a string
    """
    api_key = settings.api_key
    
    if not api_key:
        raise ValueError("GIVENERGY_API_KEY not set in environment variables")

    url = 'https://api.givenergy.cloud/v1/communication-device'
    
    headers = {
        'Authorization': f'Bearer {api_key}',
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }
    
    response = requests.get(url, headers=headers)
    
    if response.status_code != 200:
        raise Exception(f"Communication device API request failed with status code {response.status_code}")
    
    data = response.json()['data']
    if not data:
        raise ValueError("No communication devices found")
    
    inverter_serial_number = data[0]['inverter']['serial']
    return inverter_serial_number


def get_givenergy_data(settings: GivEnergySettings):
    """
    Fetch the latest data from the GivEnergy API and return a DataFrame.
    
    :return: DataFrame with timestamp and power_kw columns
    """
    api_key = settings.api_key
    
    if not api_key:
        raise ValueError("GIVENERGY_API_KEY not set in environment variables")

    inverter_serial_number = get_inverter_serial_number(settings)

    url = f'https://api.givenergy.cloud/v1/inverter/{inverter_serial_number}/system-data/latest'
    
    headers = {
        'Authorization': f'Bearer {api_key}',
        'Content-Type': 'application/json',
        'Accept': 'application/json'
    }
    
    response = requests.get(url, headers=headers)
    
    if response.status_code != 200:
        raise Exception(f"System data API request failed with status code {response.status_code}")
    
    data = response.json()['data']
    
    # Process the data
    timestamp = datetime.strptime(data['time'], "%Y-%m-%dT%H:%M:%SZ")
    power_kw = data['solar']['power'] / 1000  # Convert W to kW
    
    # Create DataFrame
    df = pd.DataFrame({
        'timestamp': [timestamp],
        'power_kw': [power_kw]
    })

    return df